﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Net;
using Jsdl.Domain;
using Newtonsoft.Json;

namespace Jsdl
{
	public abstract class CodeGenerator
    {
		public static CodeGenerator Create(string outputType)
		{
			var className = "Jsdl.Generators." + outputType + "CodeGenerator";
			var type = Type.GetType(className);
			if (type != null)
				return Activator.CreateInstance(type) as CodeGenerator;
			throw new Exception("Output type '" + outputType + "' could not be found!");
		}

		// options
		public bool AddGetOperationPrefix { get; set; }
		public bool DisableCodeGeneration { get; set; }

		// abstract
		public abstract string GetCode(string classes);
		public abstract string GetServiceClientClass(string body);

		public abstract string GetOperation(string operationName, Operation operation);

		public abstract string GetClass(string className, string classBody);
		public abstract string GetClassProperty(string propertyName, string typeName, ServiceType type);

		public abstract string GetTypeName(ServiceType type);
		public abstract string GetClassName(string className);

		// state
		public Service Service { get; set; }
		public string ClientClassName { get; set; }
		public string NamespaceName { get; set; }
		public List<ServiceClass> Classes { get; set; }

		public string GenerateFromUrl(string url, string clientClassName = "UnknownServiceClient", string namespaceName = "UnknownNamespace")
		{
			using (var client = new WebClient())
				return Generate(client.DownloadString(url), clientClassName, namespaceName);
		}

		public string Generate(string json, string clientClassName = "UnknownServiceClient", string namespaceName = "UnknownNamespace")
		{
			ClientClassName = clientClassName;
			NamespaceName = namespaceName; 
			Classes = new List<ServiceClass>();

			Service = JsonConvert.DeserializeObject<Service>(json);

			if (DisableCodeGeneration)
				return "";

			return GetCode(GenerateClasses())
				.Replace("\r", "")
				.Replace("\n", Environment.NewLine); 
		}

		private string GenerateClasses()
		{
			if (Service.Types != null)
			{
				foreach (var type in Service.Types)
					GenerateClass(type, "unknown");
			}
			
			GenerateServiceClientClass();

			var result = "";
			foreach (var c in Classes)
				result = result + IntendCode(c.Code);
			return result;
		}

		private void GenerateServiceClientClass()
		{
			if (Service.Operations != null)
			{
				var code = "";
				foreach (var op in Service.Operations)
				{
					var operation = op.Value;
					var operationName = op.Key;

					if (AddGetOperationPrefix && operation.Transport == OperationTransports.get)
						operationName = "Get" + ConvertName(operationName);

					// check constraints
					if (operation.Parameter != null) 
					{
						if (operation.ActualEnvelope != OperationEnvelope.json)
							throw new Exception("single parameter can only be used with envelope == json (" + operationName + ")");
						if (operation.Transport == OperationTransports.get)
							throw new Exception("single parameter can only be used with transport != get (" + operationName + ")");
						if (operation.Parameters != null && operation.Parameters.Count != 0)
							throw new Exception("single parameter cannot be used together with parameters (" + operationName + ")");
					}

					code += GetOperation(operationName, operation);
				}

				Classes.Add(new ServiceClass
				{
					Name = "_",
					Code = GetServiceClientClass(IntendCode(code))
				});
			}
		}

		public string GenerateClass(ServiceType type, string suggestedClassName)
		{
			var className = GetClassName(type.Name ?? suggestedClassName);
			Classes.Add(new ServiceClass
				{
					Name = className, 
					Code = GetClass(className, IntendCode(GenerateProperties(className, type)))
				});
			return className;
		}

		/// <summary>
		/// listType is either "array" or "dictionary"
		/// </summary>
		public abstract string GetArrayTypeName(string listType, string itemType);

		public string GenerateProperties(string className, ServiceType type)
		{
			var output = "";
			if (type.Properties != null)
			{
				foreach (var prop in type.Properties)
				{
					var propertyName = prop.Key;
					var property = prop.Value;
					var typeName = GenerateType(property, className + ConvertName(propertyName, false, true));
					output += GetClassProperty(propertyName, typeName, property);
				}
			}
			return output;
		}

		private int anonymousTypes = 0; 
		public string GenerateType(ServiceType type)
		{
			bool isPrimitiveType;
			anonymousTypes++; 
			return GenerateType(type, "Anonymous" + anonymousTypes, out isPrimitiveType);
		}

		public string GenerateType(ServiceType type, string suggestedClassName)
		{
			bool isPrimitiveType;
			return GenerateType(type, suggestedClassName, out isPrimitiveType);
		}

		public virtual string GenerateType(ServiceType type, string suggestedClassName, out bool isPrimitiveType)
		{
			if (type.Type == "object" || type.Type == null)
			{
				isPrimitiveType = false;
				return GenerateClass(type, suggestedClassName);
			}
			if (type.Type == "array" || type.Type == "dictionary")
			{
				isPrimitiveType = false;
				var itemType = GenerateType(type.Items, suggestedClassName);
				return GetArrayTypeName(type.Type, itemType);
			}
			
			var className = GetClassName(type.Type);
			if (Classes.Any(c => c.Name == className))
			{
				isPrimitiveType = false;
				return className;
			}

			isPrimitiveType = true;
			return GetTypeName(type);
		}
		
		public string ConvertName(string name, bool lowerStart = false, bool convertToSingular = false)
		{
			name = lowerStart ?
				name[0].ToString(CultureInfo.InvariantCulture).ToLower() + name.Substring(1) :
				name[0].ToString(CultureInfo.InvariantCulture).ToUpper() + name.Substring(1);

			if (convertToSingular && name.Last() == 's')
				name = name.Substring(0, name.Length - 1);

			int output;
			if (int.TryParse(name, out output))
				name = (lowerStart ? "num" : "Num") + output; 

			return name;
		}

		public string IntendCode(string text, int tabCount = 1)
		{
			var tabs = "";
			for (var i = 0; i < tabCount; i++)
				tabs += "\t";
			return tabs + text.Replace("\n", "\n" + tabs) + "\n";
		}
	}
}
