﻿using System.Collections.Generic;
using System.Linq;
using Jsdl.Domain;

namespace Jsdl.Generators
{
	public class CsharpCodeGenerator : CodeGenerator
	{
		public bool UseMyToolkit { get; set; }

		public override string GetCode(string classes)
		{
			var template =
@"using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System.Net;
[usings]
namespace [namespace]
{
[classes]
}";
			return template
				.Replace("[usings]", UseMyToolkit ? "using MyToolkit.Networking;\nusing System.Text;\n" : "using System.Net.Http;\n")
				.Replace("[classes]", classes)
				.Replace("[namespace]", NamespaceName); 
		}

		public override string GetServiceClientClass(string body)
		{
			var clientClassTemplate =
@"public partial class [name]
{
	public string BaseUrl { get; set; }
	public Type[] Types { get { return new Type[] { [types] }; } }

	public class JsdlException : Exception
	{
		public JsdlException(string code, string message, string serverStackTrace) : base(message) { Code = code; ServerStackTrace = serverStackTrace; }
		public string Code { get; private set; }
		public string ServerStackTrace { get; private set; }
	}

	[initializeRequest]

	public [name]() { }
	public [name](string baseUrl)
	{
		BaseUrl = baseUrl; 
	}

[body]	
}

";
			return clientClassTemplate
				.Replace("[body]", body)
				.Replace("[types]", Classes.Count > 0 ? "typeof(" + string.Join("), typeof(", Classes.Select(c => c.Name)) + ")" : "")
				.Replace("[initializeRequest]", UseMyToolkit ? 
					"partial void PrepareRequest(HttpGetRequest request);\n" +
					"partial void ProcessResponse(HttpGetRequest request, HttpResponse response);" : 
					"partial void PrepareRequest(HttpClient request);\n" +
					"partial void ProcessResponse(HttpClient request, HttpResponseMessage response);")
				.Replace("[name]", ClientClassName);
		}

		public override string GetOperation(string operationName, Operation operation)
		{
			var jsonInput = operation.ActualEnvelope == OperationEnvelope.json; 

			var postVariable = "";
			operationName = ConvertName(operationName);

			var returnType = "Task";
			var returnCode = "";

			// parameters
			var parameters = "";
			var parametersCode = "";

			if (jsonInput)
				parametersCode = "var parameters = new List<object>();\n";

			if (operation.Transport != OperationTransports.get)
				postVariable = "JsonConvert.SerializeObject(parameters)";

			var args = "";
			if (operation.Parameters != null) // POST / GET
			{
				foreach (var param in operation.Parameters)
				{
					var name = ConvertName(param.Name, true);
					parameters += GenerateType(param) + " _" + name + ", ";

					if (jsonInput)
						parametersCode += "parameters.Add(_" + name + ");\n";
					else if (operation.Envelope == OperationEnvelope.url)
						args += ", Uri.EscapeUriString(_" + name + ".ToString())";
					else // default => query
						parametersCode += "if (!object.Equals(_" + name + ", null)) { url += \"" +
							name + "=\" + Uri.EscapeUriString(\"\" + _" + name + ") + \"&\"; } \n";
				}
			}

			if (operation.Parameter != null) // POST only
			{
				if (jsonInput)
					parametersCode += "parameters.Add(_parameter);\n";
				else
					postVariable = "_parameter.ToJson()";
				parameters += GenerateType(operation.Parameter, operationName + "Request") + " _parameter, ";
			}

			parameters = parameters.Trim(' ').Trim(',');
			parametersCode = parametersCode.Trim('\n');

			// return value
			if (operation.Returns != null)
			{
				var isPrimitiveType = false;
				var type = GenerateType(operation.Returns, operationName + "Response", out isPrimitiveType);

				returnCode = "return JsonConvert.DeserializeObject<" + type + ">(data);";
				returnType = "Task<" + type + ">";
			}

			var throwException =
@"
		if (isError)
		{
			if (!string.IsNullOrEmpty(data))
			{
				try
				{
					var exception = (JContainer)JsonConvert.DeserializeObject(data);
					throw new JsdlException(
						exception[""code""] != null ? exception[""code""].ToString() : null, 
						exception[""message""] != null ? exception[""message""].ToString() : """", 
						exception[""stackTrace""] != null ? exception[""stackTrace""].ToString() : null);
				} catch { }
			}
			throw;
		}";

			var getOperationTemplate = !UseMyToolkit ?
@"public async [returnType] [name]Async([parameters])
{
	var url = BaseUrl + ""/[target][delim]"";
[parametersBody]
	var client = new HttpClient();
	PrepareRequest(client);
	var response = await client.GetAsync(url);
	ProcessResponse(client, response);
	var data = await response.Content.ReadAsStringAsync(); 
	var httpStatusCode = response.StatusCode;
	var isError = (int)httpStatusCode >= 400; 
	[throwException]
	[returnCode]
}

" :
@"public async [returnType] [name]Async([parameters])
{
	var url = BaseUrl + ""/[target][delim]"";
[parametersBody]
	var request = new HttpGetRequest(url);
	PrepareRequest(request);

	try
	{
		var response = await Http.GetAsync(request);
		ProcessResponse(request, response);
		var data = response.Response; 
		[returnCode]
	}
	catch (HttpStatusException ex)
	{
		var httpStatusCode = ex.Result.HttpStatusCode;
		var isError = (int)httpStatusCode >= 400; 
		var data = ex.Result.Response; 
		[throwException]
		else
			throw; 
	}
}

";

			var postOperationTemplate = !UseMyToolkit ?
@"public async [returnType] [name]Async([parameters])
{
	var url = BaseUrl + ""/[target][delim]"";

[parametersBody]
	var content = new StringContent([postVariable]);
	
	var client = new HttpClient();
	PrepareRequest(client);
	var response = await client.PostAsync(url, content);
	ProcessResponse(client, response);
	var data = await response.Content.ReadAsStringAsync(); 
	var httpStatusCode = response.StatusCode;
	var isError = (int)httpStatusCode >= 400; 
	[throwException]
	[returnCode]
}

" :
@"public async [returnType] [name]Async([parameters])
{
	var url = BaseUrl + ""/[target][delim]"";

[parametersBody]
	
	var request = new HttpPostRequest(url);
	request.RawData = Encoding.UTF8.GetBytes([postVariable]);
	PrepareRequest(request);
	try
	{
		var response = await Http.PostAsync(request);
		ProcessResponse(request, response);
		var data = response.Response;
		[returnCode]
	}
	catch (HttpStatusException ex)
	{
		var httpStatusCode = ex.Result.HttpStatusCode;
		var isError = (int)httpStatusCode >= 400; 
		var data = ex.Result.Response; 
		[throwException]
		else
			throw; 
	}
}

";
			return (operation.Transport == OperationTransports.get ? getOperationTemplate : postOperationTemplate)
				.Replace("[returnType]", returnType)
				.Replace("[throwException]", throwException)
				.Replace("[name]", operationName)
				.Replace("[delim]", operation.Target.Contains("?") ? "&" : "?")
				.Replace("[target]", !string.IsNullOrEmpty(args) ? 
					@""" + string.Format(""" + operation.Target + @"""" + args + @") + """ : 
					operation.Target)
				.Replace("[postVariable]", postVariable)
				.Replace("[parameters]", parameters)
				.Replace("[parametersBody]", IntendCode(parametersCode))
				.Replace("[returnCode]", returnCode);
		}

		protected string GenerateObjectConverter(ServiceType type)
		{
			// TODO implement required

			var ltype = type.Type.ToLower();
			if (ltype == "string")
				return "data;";
			if (ltype == "integer")
				return "long.Parse(data);";
			if (ltype == "number")
				return "decimal.Parse(data);";
			if (ltype == "boolean")
				return "bool.Parse(data);";

			return type + ".FromJson(data); ";
		}


		public override string GetClassProperty(string propertyName, string typeName, ServiceType type)
		{
			var template =
@"
private [type] [memberName];
[JsonProperty(""[jsonName]"", Required = [required])]
public [type] [propertyName]
{
	get { return [memberName]; }
	set 
	{
		if (value != [memberName])
		{
			[memberName] = value; 
			OnPropertyChanged(""[propertyName]"");
		}
	}
}
";
			return template
				.Replace("[type]", typeName)
				.Replace("[memberName]", ConvertName(propertyName, true))
				.Replace("[jsonName]", propertyName)
				.Replace("[required]", type.Required ? "Required.Always" : "Required.Default")
				.Replace("[propertyName]", ConvertName(propertyName));
		}

		public override string GetClass(string className, string classBody)
		{
			var template =
@"public partial class [name] : INotifyPropertyChanged
{[body]
	public string ToJson() 
	{
		return JsonConvert.SerializeObject(this);
	}

	public static [name] FromJson(string json)
	{
		return JsonConvert.DeserializeObject<[name]>(json);
	}
	
	public event PropertyChangedEventHandler PropertyChanged;

	protected virtual void OnPropertyChanged(string propertyName = null)
	{
		PropertyChangedEventHandler handler = PropertyChanged;
		if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
	}
}
";
			return template
				.Replace("[name]", className)
				.Replace("[body]", classBody);
		}

		public override string GetTypeName(ServiceType type)
		{
			var postfix = "";

			if (!type.Required && type.Type != "array" && type.Type != "dictionary" && type.Type != "object" && type.Type != "string")
				postfix = "?";

			if (type.Type == "number")
				return "decimal" + postfix;
			if (type.Type == "integer")
				return "long" + postfix;
			if (type.Type == "string" && type.Format == "date-time")
				return "DateTime" + (!type.Required ? "?" : "");

			return GetClassName(type.Type);
		}

		public override string GetArrayTypeName(string listType, string itemType)
		{
			if (listType == "array")
				return "ObservableCollection<" + itemType + ">";
			else
				return "Dictionary<string, " + itemType + ">";
		}

		public override string GetClassName(string className)
		{
			return ConvertName(className);
		}
	}
}
