﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.ServiceModel;
using Jsdl.Domain;
using Newtonsoft.Json;

namespace Jsdl
{
	public class EnumConverter : Newtonsoft.Json.Converters.StringEnumConverter
	{
		public override void WriteJson(JsonWriter writer, object value, JsonSerializer serializer)
		{
			if (value.GetType().IsEnum)
			{
				writer.WriteValue(value.ToString());
				return;
			}

			base.WriteJson(writer, value, serializer);
		}
	}

	public class JsdlGenerator
	{
		private readonly Type serviceType;
		private Service service; 

		public JsdlGenerator(Type serviceType)
		{
			this.serviceType = serviceType; 
		}

		public string Generate()
		{
			if (serviceType == null)
				return "service_not_found";

			service = new Service();
			foreach (var method in serviceType.GetMethods())
			{
				var operationAttribute = method.GetCustomAttribute<OperationContractAttribute>(true); 
				if (operationAttribute != null)
				{
					var name = operationAttribute.Name ?? method.Name; 

					var operation = new Operation();
					operation.Envelope = OperationEnvelope.json;
					operation.Transport = OperationTransports.post;
					operation.ContentTypeRaw = "application/json";
					operation.Target = "JsdlServices/" + serviceType.Name + "?operation=" + name;

					foreach (var param in method.GetParameters())
					{
						var paramType = param.ParameterType;

						var parameter = GetServiceType<ServiceParameter>(paramType);
						parameter.Name = param.Name;

						if (paramType.IsValueType || (paramType.IsGenericType && paramType.GetGenericTypeDefinition() != typeof(Nullable<>)))
							parameter.RequiredRaw = true;
						
						operation.Parameters.Add(parameter);
					}
					
					if (method.ReturnType != typeof(void))
						operation.Returns = GetServiceType<ServiceType>(method.ReturnType);
					service.Operations.Add(ConvertName(name), operation);
				}
			}
			return JsonConvert.SerializeObject(service, new JsonSerializerSettings
			{
				NullValueHandling = NullValueHandling.Ignore, 
				Converters = { new EnumConverter() }
			});
		}

		public T GetServiceType<T>(Type type) where T : ServiceType, new()
		{
			// value type
			var nullable = type.IsGenericType && type.GetGenericTypeDefinition() == typeof(Nullable<>);

			if (type == typeof(byte[]))
				return new T { TypeRaw = ConvertName("string") };

			if (type == typeof(DateTime?) || type == typeof(DateTime))
				return new T { TypeRaw = ConvertName("string"), Format = "date-time", RequiredRaw = !nullable };

			if (nullable)
				type = type.GenericTypeArguments[0]; 

			var name = type.Name; 
			if (type.IsValueType || type == typeof (string))
			{
				if (type == typeof(Int32) || type == typeof(Int64))
					name = "integer";
				if (type == typeof(Decimal) || type == typeof(Double) || type == typeof(float))
					name = "number";

				var result = new T { TypeRaw = ConvertName(name) };
				if (type != typeof (string))
					result.RequiredRaw = !nullable;
				return result; 
			}

			// object type
			if (service.Types.All(t => t.Name != name))
			{
				if (type.IsGenericType && (
					type.GetGenericTypeDefinition() == typeof(IEnumerable<>) || 
					type.GetGenericTypeDefinition() == typeof(List<>) || 
					type.GetGenericTypeDefinition() == typeof(ICollection<>)))
				{
					return new T
					{
						TypeRaw = "array", 
						Items = GetServiceType<ServiceType>(type.GetGenericArguments()[0])
					};
				}
				
				if (type.BaseType == typeof(Array))
				{
					return new T
					{
						TypeRaw = "array", 
						Items = GetServiceType<ServiceType>(type.GetElementType())
					};
				}

				var newType = new ServiceType();
				newType.Name = name;
				service.Types.Add(newType); // add here to avoid recursions

				newType.TypeRaw = "object";
				newType.Properties = new Dictionary<string, ServiceType>();
				foreach (var prop in type.GetProperties())
				{
					if (prop.GetCustomAttribute<JsonIgnoreAttribute>() == null)
					{
						var propertyName = prop.Name;

						var jsonAttribute = prop.GetCustomAttribute<JsonPropertyAttribute>();
						if (jsonAttribute != null && !string.IsNullOrEmpty(jsonAttribute.PropertyName))
							propertyName = jsonAttribute.PropertyName;

						newType.Properties.Add(propertyName, GetServiceType<ServiceType>(prop.PropertyType));
					}
				}
			}

			return new T { TypeRaw = name };
		}

		public string ConvertName(string name, bool upper = false)
		{
			if (upper)
				return name.Substring(0, 1).ToUpper() + name.Substring(1);
			return name.Substring(0, 1).ToLower() + name.Substring(1);
		}
	}
}
